
#include "pyodbc.h"
#include "pyodbcmodule.h"
#include "params.h"
#include "cursor.h"
#include "connection.h"
#include "buffer.h"
#include "wrapper.h"

enum
{
    // The largest binary parameter size we'll try to bind.  If a buffer is larger than this, we'll use SQLPutData.
    MAX_VARBINARY_BUFFER = 4000,
};

inline Connection*
GetConnection(Cursor* cursor)
{
    return (Connection*)cursor->cnxn;
}

static bool CacheParamInfo(HDBC hdbc, HSTMT hStmt, int& count, ParamInfo*& infos);
static int GetParamBufferSize(PyObject* param, Py_ssize_t iParam);
static bool BindParam(Cursor* cur, int iParam, const ParamInfo& info, PyObject* param, byte** ppbParam);


void FreeParameterData(Cursor* cur)
{
    free(cur->paramdata);
    cur->paramdata = 0;
}

static
void FreeParameterInfo(Cursor* cur)
{
    // Internal function to free just the cached parameter information.  This is not used by the general cursor code
    // since this information is also freed in the less granular free_results function that clears everything.

    Py_XDECREF(cur->pPreparedSQL);
    free(cur->paraminfos);
    cur->pPreparedSQL = 0;
    cur->paraminfos   = 0;
    cur->paramcount = 0;
}


bool PrepareAndBind(Cursor* cur, PyObject* pSql, PyObject* params, bool skip_first)
{
    // We always rebind the data since we try to bind into the Python objects directly.

    FreeParameterData(cur);

    // Normalize the parameter variables.

    int params_offset = skip_first ? 1 : 0;
    int cParams       = params == 0 ? 0 : PySequence_Length(params) - params_offset;

    if (pSql == cur->pPreparedSQL)
    {
        // We've already prepared this SQL, so we don't need to do so again.  We've also cached the parameter
        // information in cur->paraminfos.

        if (cParams != cur->paramcount)
        {
            RaiseErrorV(0, ProgrammingError, "The SQL contains %d parameter markers, but %d parameters were supplied",
                        cur->paramcount, cParams);
            return false;
        }
    }
    else
    {
        FreeParameterInfo(cur);

        SQLRETURN ret;
        if (PyString_Check(pSql))
        {
            Py_BEGIN_ALLOW_THREADS
            ret = SQLPrepare(cur->hstmt, (SQLCHAR*)PyString_AS_STRING(pSql), SQL_NTS);
            Py_END_ALLOW_THREADS
        }
        else
        {
            Py_BEGIN_ALLOW_THREADS
            ret = SQLPrepareW(cur->hstmt, (SQLWCHAR*)PyUnicode_AsUnicode(pSql), SQL_NTS);
            Py_END_ALLOW_THREADS
        }
  
        if (((Connection*)(cur->cnxn))->hdbc == SQL_NULL_HANDLE)
        {
            // The connection was closed by another thread in the ALLOW_THREADS block above.
            RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
            return false;
        }

        if (!SQL_SUCCEEDED(ret))
        {
            RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
            return false;
        }
                
        if (!CacheParamInfo(GetConnection(cur)->hdbc, cur->hstmt, cur->paramcount, cur->paraminfos))
            return false;

        cur->pPreparedSQL = pSql;
        Py_INCREF(cur->pPreparedSQL);
    }
        
    // Calculate the amount of memory we need for param_buffer.  We can't allow it to reallocate on the fly since
    // we will bind directly into its memory.  (We only use a vector so its destructor will free the memory.)
    // We'll set aside one SQLINTEGER for each column to be used as the StrLen_or_IndPtr.

    int cb = 0;

    for (Py_ssize_t i = 0; i < cParams; i++)
    {
        PyObject* param = PySequence_GetItem(params, i + params_offset);
        if (param == 0)
            return 0;

        int cbT = GetParamBufferSize(param, i + 1) + sizeof(SQLINTEGER); // +1 to map to ODBC one-based index

        Py_DECREF(param);

        if (cbT < 0)
            return 0;

        cb += cbT;
    }

    cur->paramdata = (byte*)malloc(cb);
    if (cur->paramdata == 0)
    {
        PyErr_NoMemory();
        return false;
    }

    // Bind each parameter.  If possible, items will be bound directly into the Python object.  Otherwise,
    // param_buffer will be used and ibNext will be updated.

    byte* pbParam = cur->paramdata;

    for (Py_ssize_t i = 0; i < cParams; i++)
    {
        PyObject* param = PySequence_GetItem(params, i + params_offset);

        if (!BindParam(cur, i + 1, cur->paraminfos[i], param, &pbParam))
        {
            Py_DECREF(param);
            free(cur->paramdata);
            cur->paramdata = 0;
            return false;
        }

        Py_DECREF(param);
    }

    return true;
}


static bool CacheParamInfo(HDBC hdbc, HSTMT hstmt, int& count, ParamInfo*& infos)
{
    // Called after a SQL statement is prepared to cache the number of parameters and some information about each.
    //
    // If successful, true is returned.  Otherwise, the appropriate exception will be registered with the Python system and
    // false is returned.
    //
    // count
    //   [out] The number of parameters in the statement will be returned here.  If the function fails, this will be set to
    //   zero.
    //
    // infos
    //   [out] If successful, an array of ParamInfo structures is returned here.  The array is allocated with malloc and
    //   will have one entry per parameter (will have `count` entries).  The caller is responsible for freeing this.  If
    //   the function fails, this will be set to zero.
    //

    count = 0;
    infos = 0;

    SQLSMALLINT cT;
    SQLRETURN ret;
    Py_BEGIN_ALLOW_THREADS
    ret = SQLNumParams(hstmt, &cT);
    Py_END_ALLOW_THREADS

    if (!SQL_SUCCEEDED(ret))
    {
        RaiseErrorFromHandle(hdbc, hstmt);
        return false;
    }
    
    ParamInfo* pT = (ParamInfo*)malloc(sizeof(ParamInfo) * cT);
    if (pT == 0)
    {
        PyErr_NoMemory();
        return false;
    }

    for (SQLSMALLINT i = 0; i < cT; i++)
    {
        SQLSMALLINT Nullable;
        Py_BEGIN_ALLOW_THREADS
        ret = SQLDescribeParam(hstmt, (SQLUSMALLINT)(i + 1), &pT[i].sql_type, &pT[i].column_size, &pT[i].decimal_digits, &Nullable);
        Py_END_ALLOW_THREADS

        if (!SQL_SUCCEEDED(ret))
        {
            free(pT);
            RaiseErrorFromHandle(hdbc, hstmt);
            return false;
        }
    }
        
    count = (int)cT;
    infos = pT;
    return true;
}



static int
GetParamBufferSize(PyObject* param, Py_ssize_t iParam)
{
    // Returns the size in bytes needed to hold the parameter in a format for binding, used to allocate the parameter
    // buffer.  (The value is not passed to ODBC.  Values passed to ODBC are in BindParam.)
    //
    // If we can bind directly into the Python object (e.g., using PyString_AsString), zero is returned since no extra
    // memory is required.  If the data will be provided at execution time (e.g. SQL_DATA_AT_EXEC), zero is returned
    // since the parameter value is not stored at all.  If the data type is not recognized, -1 is returned.

    if (param == Py_None)
        return 0;

    if (PyString_Check(param) || PyUnicode_Check(param))
        return 0;

    if (param == Py_True || param == Py_False)
        return 1;

    if (PyInt_Check(param))
        return sizeof(long int);

    if (PyLong_Check(param))
        return sizeof(INT64);

    if (PyFloat_Check(param))
        return sizeof(double);

    if (PyDecimal_Check(param))
    {
        // There isn't an efficient way of getting the precision, but it's there and it's obvious.

        Object digits = PyObject_GetAttrString(param, "_int");
        if (digits)
        {
            return PySequence_Length(digits) + 3;  // sign, decimal, null
        }
        
        // _int doesn't exist any more?
        return 42;
    }
    
    if (PyBuffer_Check(param))
    {
        // If the buffer has a single segment, we can bind directly to it, so we need 0 bytes.  Otherwise, we'll use
        // SQL_DATA_AT_EXEC, so we still need 0 bytes.
        return 0;
    }

    if (PyDateTime_Check(param))
        return sizeof(TIMESTAMP_STRUCT);

    if (PyDate_Check(param))
        return sizeof(DATE_STRUCT);

    if (PyTime_Check(param))
        return sizeof(TIME_STRUCT);

    RaiseErrorV("HY105", ProgrammingError, "Invalid parameter type.  param-index=%zd param-type=%s", iParam, param->ob_type->tp_name);

    return -1;
}

static bool
BindParam(Cursor* cur, int iParam, const ParamInfo& info, PyObject* param, byte** ppbParam)
{
    // Called to bind a single parameter.
    //
    // iParam
    //   The one-based index of the parameter being bound.
    //
    // param
    //   The parameter to bind.
    //
    // ppbParam
    //   On entry, *ppbParam points to the memory available for binding the current parameter.  It should be
    //   incremented by the amount of memory used.
    //
    //   Each parameter saves room for a length-indicator.  If the Python object is not in a format that we can bind to
    //   directly, the memory immediately after the length indicator is used to copy the parameter data in a usable
    //   format.
    //
    //   The memory used is determined by the type of PyObject.  The total required is calculated, a buffer is
    //   allocated, and passed repeatedly to this function.  It is essential that the amount pre-calculated (from
    //   GetParamBufferSize) match the amount used by this function.  Any changes to either function must be
    //   coordinated.  (It might be wise to build a table.  I would do a lot more with assertions, but building a debug
    //   version of Python is a real pain.  It would be great if the Python for Windows team provided a pre-built
    //   version.)

    // When binding, ODBC requires 2 values: the column size and the buffer size.  The column size is related to the
    // destination (the SQL type of the column being written to) and the buffer size refers to the source (the size of
    // the C data being written).  If you send the wrong column size, data may be truncated.  For example, if you send
    // sizeof(TIMESTAMP_STRUCT) as the column size when writing a timestamp, it will be rounded to the nearest minute
    // because that is the precision that would fit into a string of that size.

    // Every parameter reserves space for a length-indicator.  Either set *pcbData to the actual input data length or
    // set pcbData to zero (not *pcbData) if you have a fixed-length parameter and don't need it.

    SQLINTEGER* pcbData = (SQLINTEGER*)*ppbParam;
    *ppbParam += sizeof(SQLINTEGER);

    // (I've made the parameter a pointer-to-a-pointer (ergo, the "pp") so that it is obvious at the call-site that we
    // are modifying it (&p).  Here we save a pointer into the buffer which we can compare to pbValue later to see if
    // we bound into the buffer (pbValue == pbParam) or bound directly into `param` (pbValue != pbParam).
    //
    // (This const means that the data the pointer points to is not const, you can change *pbParam, but the actual
    // pointer itself is.  We will be comparing the address to pbValue, not the contents.)

    byte* const pbParam = *ppbParam;

    SQLSMALLINT ctype     = 0;
    SQLSMALLINT sqltype   = info.sql_type;
    SQLUINTEGER colsize   = info.column_size;
    SQLSMALLINT decdigits = 0;
    SQLPOINTER  pbValue   = 0; // Set to the data to bind, either into `param` or set to pbParam.
    SQLINTEGER  bufsize   = info.column_size;

    if (param == Py_None)
    {
        sqltype = info.sql_type;
        pbValue = 0;
        bufsize = 0;
        *pcbData = SQL_NULL_DATA;

        switch (sqltype)
        {
        case SQL_BINARY:
        case SQL_VARBINARY:
        case SQL_LONGVARBINARY:
            ctype = SQL_C_BINARY;
            break;

        case SQL_WCHAR:
        case SQL_WVARCHAR:
        case SQL_WLONGVARCHAR:
            ctype = SQL_C_WCHAR;
            break;

        default:
            ctype = SQL_C_CHAR;
            break;
        }
    }
    else if (PyString_Check(param))
    {
        ctype    = SQL_C_CHAR;
        pbValue  = PyString_AS_STRING(param);
        *pcbData = (SQLINTEGER)PyString_GET_SIZE(param);
    }
    else if (PyUnicode_Check(param))
    {
        ctype    = SQL_C_WCHAR;
        pbValue  = PyUnicode_AsUnicode(param);
        *pcbData = (SQLINTEGER)PyUnicode_GET_SIZE(param) * sizeof(wchar_t);
    }
    else if (param == Py_True || param == Py_False)
    {
        *pbParam = (unsigned char)(param == Py_True ? 1 : 0);

        ctype   = SQL_C_BIT;
        pbValue = pbParam;
        bufsize = 1;
        pcbData = 0;
    }
    else if (PyDateTime_Check(param))
    {
        TIMESTAMP_STRUCT* value = (TIMESTAMP_STRUCT*)pbParam;

        value->year     = (SQLSMALLINT) PyDateTime_GET_YEAR(param);
        value->month    = (SQLUSMALLINT)PyDateTime_GET_MONTH(param);
        value->day      = (SQLUSMALLINT)PyDateTime_GET_DAY(param);
        value->hour     = (SQLUSMALLINT)PyDateTime_DATE_GET_HOUR(param);
        value->minute   = (SQLUSMALLINT)PyDateTime_DATE_GET_MINUTE(param);
        value->second   = (SQLUSMALLINT)PyDateTime_DATE_GET_SECOND(param);
        value->fraction = (SQLUINTEGER)(PyDateTime_DATE_GET_MICROSECOND(param) * 1000); // 1000 == micro -> nano

        ctype   = SQL_C_TIMESTAMP;
        pbValue = pbParam;
        bufsize = sizeof(TIMESTAMP_STRUCT);
        pcbData = 0;
    }
    else if (PyDate_Check(param))
    {
        DATE_STRUCT* value = (DATE_STRUCT*)pbParam;
        value->year  = (SQLSMALLINT) PyDateTime_GET_YEAR(param);
        value->month = (SQLUSMALLINT)PyDateTime_GET_MONTH(param);
        value->day   = (SQLUSMALLINT)PyDateTime_GET_DAY(param);

        ctype   = SQL_C_DATE;
        pbValue = pbParam;
        bufsize = sizeof(DATE_STRUCT);
        pcbData = 0;
    }
    else if (PyTime_Check(param))
    {
        TIME_STRUCT* value = (TIME_STRUCT*)pbParam;
        value->hour   = (SQLUSMALLINT)PyDateTime_TIME_GET_HOUR(param);
        value->minute = (SQLUSMALLINT)PyDateTime_TIME_GET_MINUTE(param);
        value->second = (SQLUSMALLINT)PyDateTime_TIME_GET_SECOND(param);

        ctype   = SQL_C_TIME;
        pbValue = pbParam;
        bufsize = sizeof(TIME_STRUCT);
        pcbData = 0;
    }
    else if (PyInt_Check(param))
    {
        long* value = (long*)pbParam;

        *value = PyInt_AsLong(param);

        ctype   = SQL_C_LONG;
        pbValue = pbParam;
        bufsize = sizeof(long);
        pcbData = 0;
    }
    else if (PyLong_Check(param))
    {
        INT64* value = (INT64*)pbParam;

        *value = PyLong_AsLongLong(param);

        ctype   = SQL_C_SBIGINT;
        pbValue = pbParam;
        bufsize = sizeof(INT64);
        pcbData = 0;
    }
    else if (PyFloat_Check(param))
    {
        double* value = (double*)pbParam;

        *value = PyFloat_AsDouble(param);

        ctype   = SQL_C_DOUBLE;
        pbValue = pbParam;
        bufsize = sizeof(double);
        pcbData = 0;
    }
    else if (PyDecimal_Check(param))
    {
        // GetParamBufferSize reserved room for all the digits, a sign, a decimal point, and a NULL.

        Object str = PyObject_CallMethod(param, "__str__", 0);
        if (!str)
            return false;
         
        *pcbData = PyString_GET_SIZE(str.Get());
         
        ctype = SQL_C_CHAR;
        pbValue = pbParam;
        memcpy(pbValue, PyString_AS_STRING(str.Get()), *pcbData + 1);
        bufsize = info.column_size + 3;
        decdigits = info.decimal_digits;

        pcbData   = 0;
    }
    else if (PyBuffer_Check(param))
    {
        const char* pb;
        int cb = PyBuffer_GetMemory(param, &pb);

        if (cb != -1 && cb <= MAX_VARBINARY_BUFFER)
        {
            // There is one segment, so we can bind directly into the buffer object.

            ctype   = SQL_C_BINARY;
            sqltype = SQL_VARBINARY;
    
            pbValue  = (SQLPOINTER)pb;
            bufsize  = cb;
            colsize  = cb;
            *pcbData = cb;
        }
        else
        {
            // There are multiple segments, so we'll provide the data at execution time.  Pass the PyObject pointer as
            // the parameter value which will be pased back to us when the data is needed.  (If we release threads, we
            // need to up the refcount!)

            ctype   = SQL_C_BINARY;
            sqltype = SQL_LONGVARBINARY;

            pbValue  = param;
            colsize  = PyBuffer_Size(param);
            bufsize  = sizeof(PyObject*); // How big is pbValue; ODBC copies it and gives it back in SQLParamData
            *pcbData = SQL_LEN_DATA_AT_EXEC(PyBuffer_Size(param));
        }
    }
    else
    {
        RaiseErrorV("HY097", NotSupportedError, "Python type %s not supported.  param=%d", param->ob_type->tp_name, iParam);
        return false;
    }

    SQLRETURN ret = -1;
    Py_BEGIN_ALLOW_THREADS
    ret = SQLBindParameter(cur->hstmt, (USHORT)iParam, SQL_PARAM_INPUT, ctype, sqltype, colsize, decdigits, pbValue, bufsize, pcbData);
    Py_END_ALLOW_THREADS

    if (GetConnection(cur)->hdbc == SQL_NULL_HANDLE)
    {
        // The connection was closed by another thread in the ALLOW_THREADS block above.
        RaiseErrorV(0, ProgrammingError, "The cursor's connection was closed.");
        return false;
    }

    if (!SQL_SUCCEEDED(ret))
    {
        RaiseErrorFromHandle(GetConnection(cur)->hdbc, cur->hstmt);
        return false;
    }

    if (pbValue == pbParam)
    {
        // We are using the passed in buffer to bind; skip past the amount of buffer we used.
        *ppbParam += bufsize;
    }

    return true;
}

